/* ARCADEmini USB driver for Windows
 * version 1.1.1 date 2005-02-13
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <wdm.h>
#include <stdarg.h>
#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "AmUsb.h"



//global constants concerning the device
#define AmUsbMaximumTransferSize 4096 //some high value for the maximum transfer size of the USB pipes
#define AmUsbTimeoutTime -200000 //timeout time for communication steps is 20ms relative time
#define AmUsbIdVendor 0x6666 //vendor id: prototype device
#define AmUsbIdProduct 0x416D //product id: 'Am' ARCADEmini
//global constants concerning the driver
#define AmUsbVerMajor 1
#define AmUsbVerMinor 1
#define AmUsbVerRevision 1
#define AmUsbVerYear 2005
#define AmUsbVerMonth 2
#define AmUsbVerDay 13
//global constants concerning symlinks
#define AmUsbSymLinkNoMax 16 //maximum number of devices attached at same time
#define AmUsbSymLinkNamePrefix "\\DosDevices\\AmUsb" //name of the symlinks to the devices without the number



//debug-stuff
#ifdef _DEBUG
#define DBGLVL_OFF 0
#define DBGLVL_MINIMUM 1
#define DBGLVL_DEFAULT 2
#define DBGLVL_MEDIUM 3
#define DBGLVL_HIGH 4
#define DBGLVL_MAXIMUM 5
#define AmUsbDbgLvl DBGLVL_OFF
#undef AmUsbTimeoutTime //increase timeout time when debug logging is active
#define AmUsbTimeoutTime (-1000000 * (AmUsbDbgLvl + 1))
#define AmUsbAssert( Cond ) ASSERT( Cond )
#define AmUsbTrap( DbgLvl ) AmUsbTrapCond( DbgLvl, TRUE )
#define AmUsbTrapCond( DbgLvl, Cond ) { if( (AmUsbDbgLvl >= (DbgLvl)) && (Cond) ) DbgBreakPoint( ); }
#define AmUsbDebug( DbgLvl, Args ) AmUsbDebugCond( DbgLvl, TRUE, Args )
#define AmUsbDebugCond( DbgLvl, Cond, Args ) { if( (AmUsbDbgLvl >= (DbgLvl)) && (Cond) ) DbgPrint Args; }
#define AmUsbStringForDevState( DevState ) AmUsbDbgStrTabDevicePowerState[DevState]
#define AmUsbStringForSysState( SysState ) AmUsbDbgStrTabSystemPowerState[SysState]
#define AmUsbStringForPnpMnFunc( MinorFunc ) AmUsbDbgStrTabPnpMnFuncDesc[MinorFunc]
#define AmUsbStringForIrpMjFunc( MajorFunc ) AmUsbDbgStrTabIrpMajFuncDesc[MajorFunc]
#else
#define AmUsbAssert( Cond )
#define AmUsbTrap( DbgLvl )
#define AmUsbTrapCond( DbgLvl, Cond )
#define AmUsbDebug( DbgLvl, Args )
#define AmUsbDebugCond( DbgLvl, Cond, Args )
#endif
//debug-string-tables
#ifdef _DEBUG
static const PCHAR AmUsbDbgStrTabIrpMajFuncDesc[] =
{
	"IRP_MJ_CREATE",
	"IRP_MJ_CREATE_NAMED_PIPE",
	"IRP_MJ_CLOSE",
	"IRP_MJ_READ",
	"IRP_MJ_WRITE",
	"IRP_MJ_QUERY_INFORMATION",
	"IRP_MJ_SET_INFORMATION",
	"IRP_MJ_QUERY_EA",
	"IRP_MJ_SET_EA",
	"IRP_MJ_FLUSH_BUFFERS",
	"IRP_MJ_QUERY_VOLUME_INFORMATION",
	"IRP_MJ_SET_VOLUME_INFORMATION",
	"IRP_MJ_DIRECTORY_CONTROL",
	"IRP_MJ_FILE_SYSTEM_CONTROL",
	"IRP_MJ_DEVICE_CONTROL",
	"IRP_MJ_INTERNAL_DEVICE_CONTROL",
	"IRP_MJ_SHUTDOWN",
	"IRP_MJ_LOCK_CONTROL",
	"IRP_MJ_CLEANUP",
	"IRP_MJ_CREATE_MAILSLOT",
	"IRP_MJ_QUERY_SECURITY",
	"IRP_MJ_SET_SECURITY",
	"IRP_MJ_POWER",          
	"IRP_MJ_SYSTEM_CONTROL", 
	"IRP_MJ_DEVICE_CHANGE",  
	"IRP_MJ_QUERY_QUOTA",    
	"IRP_MJ_SET_QUOTA",      
	"IRP_MJ_PNP"            
};
static const PCHAR AmUsbDbgStrTabPnpMnFuncDesc[] =
{
	"IRP_MN_START_DEVICE",
	"IRP_MN_QUERY_REMOVE_DEVICE",
	"IRP_MN_REMOVE_DEVICE",
	"IRP_MN_CANCEL_REMOVE_DEVICE",
	"IRP_MN_STOP_DEVICE",
	"IRP_MN_QUERY_STOP_DEVICE",
	"IRP_MN_CANCEL_STOP_DEVICE",
	"IRP_MN_QUERY_DEVICE_RELATIONS",
	"IRP_MN_QUERY_INTERFACE",
	"IRP_MN_QUERY_CAPABILITIES",
	"IRP_MN_QUERY_RESOURCES",
	"IRP_MN_QUERY_RESOURCE_REQUIREMENTS",
	"IRP_MN_QUERY_DEVICE_TEXT",
	"IRP_MN_FILTER_RESOURCE_REQUIREMENTS",
	"IRP_MN_READ_CONFIG",
	"IRP_MN_WRITE_CONFIG",
	"IRP_MN_EJECT",
	"IRP_MN_SET_LOCK",
	"IRP_MN_QUERY_ID",
	"IRP_MN_QUERY_PNP_DEVICE_STATE",
	"IRP_MN_QUERY_BUS_INFORMATION",
	"IRP_MN_DEVICE_USAGE_NOTIFICATION",
	"IRP_MN_SURPRISE_REMOVAL"
};
static const PCHAR AmUsbDbgStrTabSystemPowerState[] = 
{
	"PowerSystemUnspecified",
	"PowerSystemWorking",
	"PowerSystemSleeping1",
	"PowerSystemSleeping2",
	"PowerSystemSleeping3",
	"PowerSystemHibernate",
	"PowerSystemShutdown",
	"PowerSystemMaximum"
};
static const PCHAR AmUsbDbgStrTabDevicePowerState[] = 
{
	"PowerDeviceUnspecified",
	"PowerDeviceD0",
	"PowerDeviceD1",
	"PowerDeviceD2",
	"PowerDeviceD3",
	"PowerDeviceMaximum"
};
#endif



//minumum and maximum
#define min( a, b ) ((a) < (b) ? (a) : (b))
#define max( a, b ) ((a) > (b) ? (a) : (b))

//number of elements in an array
#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))

//byte order
#define ntohs( net ) ( (unsigned short)((unsigned char *)&(unsigned short)(net))[0] << 8 \
                     | (unsigned short)((unsigned char *)&(unsigned short)(net))[1] )
#define ntohl( net ) ( (unsigned long)((unsigned char *)&(unsigned long)(net))[0] << 24 \
                     | (unsigned long)((unsigned char *)&(unsigned long)(net))[1] << 16 \
                     | (unsigned long)((unsigned char *)&(unsigned long)(net))[2] << 8 \
                     | (unsigned long)((unsigned char *)&(unsigned long)(net))[3] )
//this is a terrible hack:
//  - htons/htonl must swap bytes if and olny if ntohs/nthol swaps bytes
//  - so we can define htons/htonl to be the same
#define htons ntohs
#define htonl ntohl



//a communication step
typedef struct tAmUsbCommStep
{
  int OutLen; //length of output buffer
  unsigned char * pOutBuffer; //buffer with data to send
  int InLen; //length of input buffer
  unsigned char * pFilterBuffer; //buffer with filter for data to receive (may be NULL)
  unsigned char * pMaskBuffer; //buffer with mask for data to receive (may be NULL)
  unsigned char * pInBuffer; //buffer for received data (may be NULL)
} stAmUsbCommStep;

//a communication block
typedef struct tAmUsbCommBlock
{
  struct tAmUsbCommBlock * pNext, * pPrev; //doubly linked list
  int Linked; //flag if currently linked in doubly linked list
  int StepCnt; //number of communication steps
  struct tAmUsbCommStep * Steps; //array with communication steps (GFP_KERNEL)
  int OutStep; //next step to output
  int InStep; //next step to input
  int Timeout; //set to != 0 if timed out
  PRKEVENT pWaitEvent; //event to signal after completion of this block
                       //(this pointer is only a copy and is not used to free the event)
                       //(may be NULL)
  int FreeFlag; //if != 0, this communication block is automatically freed after processing
} stAmUsbCommBlock;



//type for private data of a file
typedef struct tAmUsbFilePriv
{
	struct tAmUsbDev * pDev; //pointer to associated device
	struct tAmUsbFilePriv * pNext, * pPrev; //pointer to next and previous file that have opened same device

	int Mode; //the mode of this file
} stAmUsbFilePriv;

//type for private data of a device
typedef struct tAmUsbDev
{
	FAST_MUTEX FileLock; //used to syncronize file operations on this device
	KSPIN_LOCK Lock; //used to lock this structure

	struct _DEVICE_EXTENSION * pDeviceExtension; //pointer to device extension to which this private device data belongs

	int Unload; //set if device is being unloaded (e.g. to prevent scheduling of DPCs)

	int Removed; //set if device is removed but still opened by someone

	unsigned char OutEndpointAddr; //address of the bulk out endpoint
	int OutMaxPacketSize; //maximum size of packets for bulk out endpoint
	USBD_PIPE_HANDLE OutPipeHandle; //handle of the out pipe
	unsigned long OutActive; //boolean flag, if bulk out is active
	PIRP pOutIrp; //pointer to bulk out IRP
	PURB pOutUrb; //pointer to bulk out URB
	unsigned char * pOutBuffer; //pointer to bulk out buffer
	int OutBufferSize; //size of bulk out buffer
	unsigned long OutCancelled; //boolean flag, if bulk out is beeing cancelled (see IRP-CANCEL-PROCEDURE below)

	unsigned char InEndpointAddr; //address of the bulk in endpoint
	int InMaxPacketSize; //maximum size of packets for bulk in endpoint
	USBD_PIPE_HANDLE InPipeHandle; //handle of the in pipe
	unsigned long InActive; //boolean flag, if bulk in is active
	PIRP pInIrp; //pointer to bulk in IRP
	PURB pInUrb; //pointer to bulk in URB
	unsigned char * pInBuffer; //pointer to bulk in buffer
	int InBufferSize; //size of bulk in buffer
	unsigned long InCancelled; //boolean flag, if bulk in is beeing cancelled (see IRP-CANCEL-PROCEDURE below)

	KDPC DpcObject; //DPC object (used to queue some tasks that can not be done in interrupt)
	//windows does not know atomic bit operations,
	//so we have to reserve an entire long for each task flag
	long TaskInStart; //start USB bulk input
	long TaskCommBlockSend; //try to send next step of communication block
	long TaskCommBlockRecv; //post-process received steps of communication block

	int OpenCnt; //numer of times this file is opened at the moment
	struct tAmUsbFilePriv FilePrivList; //anchor of list with files that have opened this device

	struct tAmUsbCommBlock CommBlockList; //anchor of list of communication blocks

	int TimeoutTimerActive; //if != 0 timeout timer is currently running
	KTIMER TimeoutTimer; //timer for timeout of USB bulk communication
	KDPC TimeoutDpc; //DPC to be called by TimeoutTimer

	int Mode; //the mode to use for new files
} stAmUsbDev;



/* IRP-CANCEL-PROCEDURE
 * found in the newsgroups
 *
 * I think it's safe to cancel an IPR that you've allocated with
 * IoBuildAsynchronousFsdRequest or IoAllocateIrp using the following
 * rather strange technique. [The entire rest of this message needs a Nerd
 * Alert.]
 *
 * 1. Your device extension has two extra fields:
 *
 * PIRP TheIrp;  // this is the asynchronous IRP you might want to cancel later
 * ULONG cancelled; // a flag
 *
 * 2. Your completion routine looks like this:
 *
 * NTSTATUS CompletionRoutine(PDEVICE_OBJECT junk, PIRP Irp, PDEVICE_EXTENSION pdx)
 * {
 *   if (InterlockedExchangePointer(&pdx->TheIrp, NULL)
 *    || InterlockedExchange(&pdx->cancelled, 1))
 *     CleanupTheIrp(Irp);
 *   return STATUS_MORE_PROCESSING_REQUIRED;
 * }
 *
 * 3. Before launching the IRP, be sure to save TheIrp and to set
 * "cancelled" to zero.
 *
 * 4. If you ever decide to cancel the IRP, use this logic:
 *
 * PIRP Irp = (PIRP) InterlockedExchangePointer(&pdx->TheIrp, NULL);
 * if (Irp)
 * {
 *   IoCancelIrp(Irp);
 *   if (InterlockedExchange(&pdx->cancelled, 1))
 *     CleanupTheIrp(Irp);
 *   }
 *
 * 5. CleanupTheIrp is another routine you write. It calls things like
 * IoFreeMdl to cleanup any auxiliary data structures associated with the
 * IRP and ends up calling IoFreeIrp.
 *
 * Here's how this works. In a normal case, there aren't any races. If
 * someone completes the IRP, your completion routine finds TheIrp non-NULL
 * and releases the IRP. You'll never call IoCancelIrp because TheIrp will
 * already be NULL at the time you make the decision.
 *
 * The other halfway normal case would be when you decide to cancel the
 * IRP. In this case, you find TheIrp non-NULL and call IoCancelIrp.
 * Whoever owns the IRP has a cancel routine that will call
 * IoCompleteRequest, which calls your completion routine. It finds TheIrp
 * NULL and therefore performs the further test on "cancelled", which is
 * still zero. CleanupTheIrp then runs and releases the IRP. When
 * IoCancelIrp returns, you see that "cancelled" is now one and you stop.
 *
 * Now for the races... Let's say you decide (one one CPU) you want to
 * cancel the IRP, but someone (on another CPU) has already started the
 * completion process. If the completion process gets past the
 * InterlockedExchangePointer in the completion routine, your cancel logic
 * will find the pointer NULL and won't do anything. The completion routine
 * won't test "cancelled" in this situation, and it will release the IRP.
 *
 * But suppose it happens in the other order, such that you really do call
 * IoCancelIrp. If the completion routine runs before IoCancelIrp returns,
 * it will find TheIrp to be NULL and "cancelled" to be zero. It will skip
 * the cleanup step in this case. As soon as IoCancelIrp returns, you'll
 * see that "cancelled" is now "1", and you'll call CleanupTheIrp.
 *
 * In the scenario I described in the preceding paragraph, it's also
 * possible to IoCancelIrp to return before the completion routine gets to
 * the point of testing "cancelled". In this case, your cancellation path
 * will see "cancelled" still 0 and stop. The completion routine will now
 * see that "cancelled" is now "1" and cleanup the IRP.
 */


 
//a structure representing the instance information associated with this particular device.
typedef struct _DEVICE_EXTENSION
{
	PDEVICE_OBJECT pTopOfStackDeviceObject; //device object we call when submitting URBs
	PDEVICE_OBJECT pPhysicalDeviceObject; //the bus driver object
	DEVICE_POWER_STATE CurrentDevicePowerState;
	USBD_CONFIGURATION_HANDLE hUsbConfiguration; //handle to config
	PUSB_CONFIGURATION_DESCRIPTOR pUsbConfigurationDescriptor; //config the device is in
	PUSB_DEVICE_DESCRIPTOR pUsbDeviceDescriptor; //pointer to USB device descriptor
	PUSBD_INTERFACE_INFORMATION pUsbInterface; //copy of the interface structure
	DEVICE_CAPABILITIES DeviceCapabilities; //device caps set by bus driver, might be altered by function driver
	PIRP pPowerIrp; //currently-being-handled system-requested power IRP request
	KEVENT evRemove; //set when PendingIoCount goes to 0; flags device can be removed
	KEVENT evNoPendingIo; //set when PendingIoCount goes to 1 (1st increment was on add device) - this indicates that no IO pending
	KEVENT evSelfRequestedPowerIrp; //set to signal driver-generated power request is finished
	KSPIN_LOCK IoCountSpinLock; //spinlock used to protect inc/dec IO count logic
	ULONG PendingIoCount; //incremented when device is added and any IO request is received; decremented when any io request is completed or passed on, and when device is removed
	WCHAR DeviceLinkNameBuffer[MAXIMUM_FILENAME_LENGTH]; //name buffer for our named functional device object link, the name is generated based on the driver's class GUID
	BOOLEAN DeviceRemoved; //flag set when processing IRP_MN_REMOVE_DEVICE
	BOOLEAN RemoveDeviceRequested; //flag set when driver has answered success to IRP_MN_QUERY_REMOVE_DEVICE
	BOOLEAN StopDeviceRequested; //flag set when driver has answered success to IRP_MN_QUERY_STOP_DEVICE
	BOOLEAN DeviceStarted; //flag set when device has been successfully started
	BOOLEAN EnabledForWakeup; //flag set when IRP_MN_WAIT_WAKE is received and we're in a power state where we can signal a wait
	BOOLEAN SelfPowerIrp; //used to flag that we're currently handling a self-generated power request
	ULONG PowerDownLevel; //default power state to power down to on self-suspend
	ULONG MaximumTransferSize; //default maximum transfer per staged IRP size
	unsigned long AppOpenCnt; //by how many userspace apps this device is opened at the moment
	struct tAmUsbDev * pDev; //pointer to private device data
	BOOLEAN SymLinkSet; //flag to indicate that a symlink to the device is set
	UNICODE_STRING SymLink; //symlink to device object
} DEVICE_EXTENSION, * PDEVICE_EXTENSION;



//prototypes from AmUsb.c
LONG AmUsbDecIoCount( IN PDEVICE_OBJECT pDeviceObject );
LONG AmUsbIncIoCount( IN PDEVICE_OBJECT pDeviceObject );
NTSTATUS AmUsbCreateDeviceObject( IN PDRIVER_OBJECT pDriverObject, IN PDEVICE_OBJECT pPhysicalDeviceObject, IN PDEVICE_OBJECT * ppDeviceObject );
NTSTATUS AmUsbCallUsbd( IN PDEVICE_OBJECT pDeviceObject, IN PURB pUrb );
NTSTATUS AmUsbConfigureDevice( IN PDEVICE_OBJECT pDeviceObject );
NTSTATUS AmUsbResetPipe( IN PDEVICE_OBJECT pDeviceObject, IN PUSBD_PIPE_INFORMATION pPipeInfo );
//prototypes from AmUsbIo.c
BOOLEAN AmUsbCanAcceptIoRequests( IN PDEVICE_OBJECT pDeviceObject );
NTSTATUS AmUsbRead( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp );
NTSTATUS AmUsbWrite( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp );
NTSTATUS AmUsbIoctl( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp );
NTSTATUS AmUsbClose( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp );
NTSTATUS AmUsbCreate( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp );
//prototypes from AmUsbPnp.c
NTSTATUS AmUsbIrpCompletionRoutine( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp, IN PVOID pContext );
NTSTATUS AmUsbPnp( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp );    
NTSTATUS AmUsbPnpAddDevice( IN PDRIVER_OBJECT pDriverObject, IN PDEVICE_OBJECT pPhysicalDeviceObject );
//prototypes from AmUsbPwr.c
NTSTATUS AmUsbPower( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp );    
NTSTATUS AmUsbSelfSuspendOrActivate( IN PDEVICE_OBJECT pDeviceObject, IN BOOLEAN fSuspend );
NTSTATUS AmUsbQueryCapabilities( IN PDEVICE_OBJECT pPdoDeviceObject, IN PDEVICE_CAPABILITIES pDeviceCapabilities );
//prototypes from AmUsbFunc.c
NTSTATUS AmUsbFileRead( PFILE_OBJECT pFileObject, char * pBuffer, unsigned int Count, unsigned int * pLen );
NTSTATUS AmUsbFileWrite( PFILE_OBJECT pFileObject, char * pBuffer, unsigned int Count, unsigned int * pLen );
NTSTATUS AmUsbFileIoctl( PFILE_OBJECT pFileObject, unsigned int IoctlNo, void * pBuffer, unsigned int InLen, unsigned int OutBufLen, unsigned int * pOutLen );
NTSTATUS AmUsbFileOpen( struct tAmUsbDev * pDev, PFILE_OBJECT pFileObject );
NTSTATUS AmUsbFileRelease( PFILE_OBJECT pFileObject );
int AmUsbProbe( PDEVICE_EXTENSION pDeviceExtension );
void AmUsbDisconnect( PDEVICE_EXTENSION pDeviceExtension );
